﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Drawing;
using System.Collections.Generic;
using HeuristicLab.Problems.VehicleRouting.Encodings.General;
using HeuristicLab.Problems.VehicleRouting.Variants;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Problems.VehicleRouting.ProblemInstances;

namespace HeuristicLab.Problems.VehicleRouting.Encodings.Potvin {
  [Item("PotvinEncoding", "Represents a potvin encoding of VRP solutions. It is implemented as described in Potvin, J.-Y. and Bengio, S. (1996). The Vehicle Routing Problem with Time Windows - Part II: Genetic Search. INFORMS Journal of Computing, 8:165–172.")]
  [StorableClass]
  public class PotvinEncoding : TourEncoding, ISingleDepotEncoding, ITimeWindowedEncoding, IHomogenousCapacitatedEncoding {   
    [Storable]
    public List<int> Unrouted { get; set; }

    public override IDeepCloneable Clone(HeuristicLab.Common.Cloner cloner) {
      PotvinEncoding clone = new PotvinEncoding(ProblemInstance);
      cloner.RegisterClonedObject(this, clone);
      clone.Tours = (ItemList<Tour>)cloner.Clone(this.Tours);
      clone.Unrouted = new List<int>(Unrouted);
      return clone;
    }

    public PotvinEncoding(IVRPProblemInstance instance)
      : base(instance) {
      Unrouted = new List<int>();
    }

    [StorableConstructor]
    private PotvinEncoding(bool serializing)
      : base(serializing) {
    }

    public static PotvinEncoding ConvertFrom(IVRPEncoding encoding, IVRPProblemInstance instance) {
      PotvinEncoding solution = new PotvinEncoding(instance);

      TourEncoding.ConvertFrom(encoding, solution, instance);

      return solution;
    }

    public static PotvinEncoding ConvertFrom(List<int> route, IVRPProblemInstance instance) {
      PotvinEncoding solution = new PotvinEncoding(instance);

      TourEncoding.ConvertFrom(route, solution);

      return solution;
    }

    public double GetTourLength(Tour tour) {
      double length = 0;

      if (tour.Stops.Count > 0) {
        List<int> cities = new List<int>();
        cities.Add(0);
        foreach (int city in tour.Stops) {
          cities.Add(city);
        }
        cities.Add(0);

        for (int i = 1; i < cities.Count; i++) {
          length += ProblemInstance.GetDistance(cities[i - 1], cities[i]);
        }
      }  

      return length;
    }

    public bool FindInsertionPlace(int city, int routeToAvoid, out int route, out int place) {
      route = -1;
      place = -1;
      double minDetour = 0;

      for (int tour = 0; tour < Tours.Count; tour++) {
        if (tour != routeToAvoid) {
          double length = GetTourLength(Tours[tour]);

          for (int i = 0; i <= Tours[tour].Stops.Count; i++) {
            Tours[tour].Stops.Insert(i, city);

            VRPEvaluation eval = ProblemInstance.EvaluatorParameter.Value.Evaluate(
              ProblemInstance, Tours[tour]);
            if (ProblemInstance.EvaluatorParameter.Value.Feasible(eval)) {
              double newLength = eval.Distance;

              double detour = newLength - length;

              if (route <= 0 || detour < minDetour) {
                route = tour;
                place = i;
                minDetour = detour;
              }
            }

            Tours[tour].Stops.RemoveAt(i);
          }
        }
      }

      return route >= 0 && place >= 0;
    }
  }
}
